@@ -21,9 +21,9 @@ except ImportError:
     from urllib3.response import HTTPResponse
 
 if six.PY2:
-    from urlparse import urlparse, parse_qsl
+    from urlparse import urlparse, parse_qsl, urlsplit, urlunsplit
 else:
-    from urllib.parse import urlparse, parse_qsl
+    from urllib.parse import urlparse, parse_qsl, urlsplit, urlunsplit
 
 if six.PY2:
     try:
@@ -102,12 +102,12 @@ class CallList(Sequence, Sized):
         self._calls = []
 
 
-def _ensure_url_default_path(url, match_querystring):
-    if _is_string(url) and url.count('/') == 2:
-        if match_querystring:
-            return url.replace('?', '/?', 1)
-        else:
-            return url + '/'
+def _ensure_url_default_path(url):
+    if _is_string(url):
+        url_parts = list(urlsplit(url))
+        if url_parts[2] == '':
+            url_parts[2] = '/'
+        url = urlunsplit(url_parts)
     return url
 
 
@@ -127,7 +127,8 @@ class BaseResponse(object):
         self.method = method
         self.match_querystring = match_querystring
         # ensure the url has a default path set if the url is a string
-        self.url = _ensure_url_default_path(url, match_querystring)
+        self.url = _ensure_url_default_path(url)
+        self.call_count = 0
 
     def _url_matches_strict(self, url, other):
         url_parsed = urlparse(url)
@@ -138,6 +139,10 @@ class BaseResponse(object):
 
         url_qsl = sorted(parse_qsl(url_parsed.query))
         other_qsl = sorted(parse_qsl(other_parsed.query))
+
+        if len(url_qsl) != len(other_qsl):
+            return False
+
         for (a_k, a_v), (b_k, b_v) in zip(url_qsl, other_qsl):
             if not isinstance(a_k, six.text_type):
                 a_k = a_k.decode('utf-8')
@@ -231,7 +236,7 @@ class Response(BaseResponse):
 
         return HTTPResponse(
             status=status,
-            reason=six.moves.http_client.responses[status],
+            reason=six.moves.http_client.responses.get(status),
             body=body,
             headers=headers,
             preload_content=False, )
@@ -263,7 +268,7 @@ class CallbackResponse(BaseResponse):
 
         return HTTPResponse(
             status=status,
-            reason=six.moves.http_client.responses[status],
+            reason=six.moves.http_client.responses.get(status),
             body=body,
             headers=headers,
             preload_content=False, )
@@ -291,7 +296,13 @@ class RequestsMock(object):
         self._matches = []
         self._calls.reset()
 
-    def add(self, method_or_response=None, url=None, body='', *args, **kwargs):
+    def add(self,
+            method_or_response=None,
+            url=None,
+            body='',
+            adding_headers=None,
+            *args,
+            **kwargs):
         """
         A basic request:
 
@@ -331,6 +342,9 @@ class RequestsMock(object):
             self._matches.append(method_or_response)
             return
 
+        if adding_headers is not None:
+            kwargs.setdefault('headers', adding_headers)
+
         if method_or_response:
             kwargs['method'] = method_or_response
 
@@ -377,11 +391,9 @@ class RequestsMock(object):
                 break
         else:
             return None
-
-        if self.assert_all_requests_are_fired:
-            # for each found match remove the url from the stack
-            self._matches.remove(match)
-
+        # shift matches
+        self._matches = self._matches[1:]
+        self._matches.append(match)
         return match
 
     def _on_request(self, adapter, request, **kwargs):
@@ -403,6 +415,7 @@ class RequestsMock(object):
                 request,
                 match.get_response(request), )
         except Exception as response:
+            match.call_count += 1
             self._calls.add(request, response)
             response = resp_callback(response) if resp_callback else response
             raise
@@ -418,6 +431,7 @@ class RequestsMock(object):
             pass
 
         response = resp_callback(response) if resp_callback else response
+        match.call_count += 1
         self._calls.add(request, response)
         return response
 
@@ -436,10 +450,15 @@ class RequestsMock(object):
 
     def stop(self, allow_assert=True):
         self._patcher.stop()
-        if allow_assert and self.assert_all_requests_are_fired and self._matches:
+        if not self.assert_all_requests_are_fired:
+            return
+        if not allow_assert:
+            return
+        not_called = [m for m in self._matches if m.call_count == 0]
+        if not_called:
             raise AssertionError(
                 'Not all requests have been executed {0!r}'.format([(
-                    match.method, match.url) for match in self._matches]))
+                    match.method, match.url) for match in not_called]))
 
 
 # expose default mock namespace
